uniform float width;
uniform float height;

uniform sampler3D densities;
uniform sampler2D outlets;
uniform sampler3D gradients;
uniform sampler2D transfer;

varying mat3 normMatrix;

const vec4 pl = vec4(1.0, 1.0, 1.0, 1.0);
vec4 ka = vec4(0.3, 0.3, 0.3, 1.0);
vec4 kd = vec4(0.6, 0.6, 0.6, 1.0);
const vec4 ks = vec4(1.0, 1.0, 1.0, 1.0);
const float n = 32.0;
const float k1 = 1.0;
const float k2 = 0.0;
const vec3 L = normalize(vec3(-1.0, +1.0, -1.0));


void main(void) {

	float step = 0.002;
	
	vec3 end = texture2D(outlets, vec2(gl_FragCoord[0]/width, 1-gl_FragCoord[1]/height)).stp;
	vec3 start = gl_TexCoord[0].stp;
	
	vec3 V = normalize(end-start);
	
	vec3 position = start;
	vec4 col = vec4(0.0,0.0,0.0,0.0);
	
	while (position[0] >= 0 && position[1] >= 0 && position[2] >= 0 && position[0] <= 1 && position[1] <= 1 && position[2] <= 1) {
		
		vec2 density = texture3D(densities, position).bg;
		
		float h = floor(density.y*16.0)/16.0;
		float h2 = density.y - h;
		density.y /= 16.0;
		density.x /= 16.0;
		density.x += h2*16.0;
		
		density.y = 1 - density.y;
		density.x = 1 - density.x;
		
		kd = texture2D(transfer, density);
		ka = kd * 0.5;
		
		vec3 N = normalize( texture3D(gradients, position).xyz - vec3(0.5, 0.5, 0.5) );
		vec3 H = normalize(normMatrix*V+L);
		
		vec3 c = pl * ka + pl / (k1 + k2*gl_FragDepth) * (kd * max(dot(N,L),0.0) + ks * pow( max(dot(N,H),0.0), n ));
		vec4 help = texture3D(densities, position);
		float alpha = kd.a;
		
		vec3 tmp = col.rgb * col.a + c * (1-col.a);
		col = vec4(tmp[0], tmp[1], tmp[2], col.a + alpha*(1.0-col.a) );
		
		/*if (col.a >= 1) break;*/
		
		position += V * step;
		
	}
	
	col = col * col.a;
	col.a = 1.0;

	/*gl_FragColor = vec4(end[0], end[1], end[2], 1.0);*/
	/*gl_FragColor = vec4(start[0], start[1], start[2], 1.0);*/
	gl_FragColor = col;

}